#include <iostream>
#include <vector>
#include <fstream>
#include <cmath>
using namespace std;

double CL = 0;  // on met ud=0 pour simplifier


//==============================================
//	Classe Matelement 
//	rend une matrice elementaire 2x2
//	et sa position par iK
//==============================================
class MatElement { 
public:
  int iK[2];                //les numeros des 2 sommets de l’element K
  double a [2][2] ;                                     //    les 2x2 coef
public:
  MatElement(int k,double a, vector<double> b, double c);
  };
//==============================================





//===============================================
//	Classe MatTriDia
//	qui rend 3 vecteurs a, b , c de taille n
//===============================================
class MatTriDia {
public:
  int n ;                          //    nb de lignes et/ou colonnes
  vector<double> a,b,c ;
public:
  MatTriDia(int nn);  //initialisateur
  void MiseAZero();
  void Mul(vector<double> &x);
  MatTriDia & operator+= (MatElement &Ak);
  void Dirichlet();
  void Lu();
  void Solve();
};
//===============================================

//===============================================
//	initialisateur de la classe MatTriDia
//===============================================
MatTriDia::MatTriDia(int nn){
n=nn;
a.resize(n) ;
b.resize(n) ;
c.resize(n) ;
}
//===============================================


//===============================================
//	Mise a zero des variables de MatTriDia
//	met les vecteurs a, b , c a zeros
//	presque inutile car le copilateur le faisait par defaut
//===============================================
void MatTriDia::MiseAZero() {
vector<double> a(n,0) , b(n,0) , c(n,0);
}
//===============================================

//===============================================
//	Permet de multiplier A a x
//===============================================
void MatTriDia::Mul(vector<double> &x){
vector<double> Ax(n,0);
int h1 ,h2;
h1=x.size();
h2=Ax.size();
if ( h1!=h2){ cerr<< "erreur la matrice et le vecteur ne sont pas de meme taille" <<endl;
exit(0);}
else 
{Ax[0] = b[0] * x[0] +c[0]*x[1];
for (int i=1 ; i<n-2; i++){
Ax[i]= a[i]*x[i-1] + b[i]* x[n]+c[i]*x[i+1];
}
Ax[n-1]= a[n-1]*x[n-2] + b[n-1]* x[n-1];}
}
//==============================================

//==============================================
//	Compte le nombre de double dans un fichier
//==============================================
int Nbl(const char *fichier)
{ifstream fic(fichier);
double c;
int compte=0;
while ( (fic >>c).good() ){			//lecture de chaque ligne
compte++;
}
fic.clear();
return compte;}
//==============================================

//==============================================
//	Calcul et creait la matrice elementaire
//==============================================
MatElement::MatElement( int k, double ap, vector<double> bp, double cp){
ifstream fic("pas.dat");
int t = Nbl("pas.dat");
vector<double> pas(t,0);
for (int i=0; i<(t); i++){fic>>pas[i];}

double h = (pas[k+1] - pas[k]) ;
            
a[0][0]=(ap/(h)) - (bp[0]/(2.)) + (cp*h)/3.; 
a[1][0]=-(ap/(h)) + (bp[0]/(2.)) + (cp*h)/6.;
a[0][1]=-(ap/(h)) - (bp[1]/(2.)) + (cp*h)/6. ;
a[1][1]=(ap/(h)) + (bp[1]/(2.)) + (cp*h)/3. ;
iK[0]=k;
iK[1]=k+1;

//permet d avoir la derniere matrice Ak calculer. Bien pour tester
//ofstream ff("Ak.dat");
//ff<<  (ap/(h)) - (bp[1]/(2.*h)) + (cp*h)/3.<< "\t" << -(ap/(h*h)) - (bp[2]/(2.*h)) + (cp*h)/6.<<endl;
//ff<< -(ap/(h)) - (bp[1]/(2.*h)) + (cp*h)/6.<< "\t" <<  (ap/(h*h)) - (bp[2]/(2.*h)) + (cp*h)/3.<<endl;
}
//============================================== 


//==============================================
//	Ajoute Ak a la matrice A 
//	en position k
//==============================================
MatTriDia & MatTriDia::operator+=(MatElement &Ak)
{
b[Ak.iK[0]]+=Ak.a[0][0];
c[Ak.iK[0]]+=Ak.a[0][1];
a[Ak.iK[1]]+=Ak.a[1][0];
b[Ak.iK[1]]+=Ak.a[1][1];
}
//==============================================


//==============================================
//	Met les conditions de Dirichlet
//	nul aux bords
//==============================================
void MatTriDia::Dirichlet(){
b[0]=1; 		//ou b[0]=b[0]*tgv   dans ce cas la on n a pas besoin de mettre c a zero
c[0]=0;
b[n-1]=1;		//ou b[n-1]=b[n-1]*tgv
a[n-1]=0;}
//==============================================

//==============================================
//	Transforme A en LU
//	pour permettre sa resolution
//==============================================
void MatTriDia::Lu(){
b[0]=b[0];
c[0]=c[0]/b[0];

for(int i=1;i<n;i++){
	b[i]=b[i]-(a[i]*c[i-1]);
	c[i]=c[i]/b[i];}
}
//==============================================

//==============================================
//	Resoud LU en prenant F:"second.dat"
//	comme second membre
//==============================================
void MatTriDia::Solve(){
ifstream fl("second.dat");
int N=Nbl("second.dat");

vector <double> f(N/2);
vector<double> pas(N/2);

for (int i=0;i<(N/2);i++){
fl >>pas[i];
fl >> f[i];}
fl.clear();

//Calcul de LY=F
vector <double> Y(N/2);
Y[0]=(f[1]+CL)/b[0];
for (int i=1;i<(N/2);i++){Y[i]=((f[i]+CL) - (a[i]*Y[i-1]))/b[i];}

//Calcul de UX=Y
vector<double> X(N/2);
X[(N/2)-1]=Y[(N/2)-1];
for (int i=(N/2)-2;i>=0;i--){X[i]=Y[i]-(c[i]*X[i+1]);}

//sauvegarde des resultats
ofstream sol("sol.dat");
for (int i=0;i<(N/2);i++){sol<<pas[i]<<"\t"<<X[i]<<endl;}
cout<< "fichier solution ecrit dans : sol.dat"<<endl;}

//==============================================

//==============================================
//	Lance tout
//==============================================
int main(){
vector <double> m(2);
m[0]=1;
m[1]=1;

//		Test valeur de iK et de Ak pour le iK donné
//	MatElement Ak(149,1,b,1);
//	cout<< Ak.iK[0]<<"\t"<<Ak.iK[1]<<endl;
//	cout<< Ak.a[0][0] <<"\t"<< Ak.a[0][1] <<endl;
//	cout<< Ak.a[1][0] <<"\t"<< Ak.a[1][1] <<endl;

int N = Nbl("pas.dat") -1;  //on a N+1 points dans pas.dat, donc N intervalles
MatTriDia p(N);
cout <<"taille pas :"<<N+1<<endl;
p.MiseAZero();

//Calcul de la matrice A
for (int i=0; i < N-1 ;i++) { MatElement ak(i,1,m,1); p+=ak;}

	ofstream ma("A.dat");
	for (int i=0; i < N ;i++){ma << p.a[i] <<"\t"<< p.b[i] <<"\t"<< p.c[i]<<endl; }
	cout<< "Matrice A ecrite dans : A.dat"<<endl;

p.Dirichlet();
p.Lu();

	ofstream mat("Alu.dat");
	for (int i=0; i < N ;i++){mat << p.a[i] <<"\t"<< p.b[i] <<"\t"<< p.c[i]<<endl; }
	cout<< "Matrice LU ecrite dans : Alu.dat"<<endl;
	cout << "Taille de Alu :"<< Nbl("Alu.dat")/3<< endl;

p.Solve();

	cout << "Taille de sol :"<< Nbl("sol.dat")/2<< endl;
	cout << "Dans gnuplot, tapez:plot \"sol.dat\" w l"<<endl;
//cout<< " Dans gnuplot, tapez :plot \"sol.dat\" w l, \"sol.dat\" using 1:3 w l "<<endl;
//cout<< " Dans gnuplot, tapez :plot \"sol.dat\" w l, \"solex.dat\" w l "<<endl;
return 0 ; }
//==============================================
