#include <fstream>
#include <iostream>
#include <cassert>
#include <algorithm>
#include <complex>


#include "gmres.hpp"

 using namespace std;
#include <time.h> 
inline double CPUtime(){
#ifdef SYSTIMES
  struct tms buf;
  if (times(&buf)!=-1)
    return ((double)buf.tms_utime+(double)buf.tms_stime)/(long) sysconf(_SC_CLK_TCK);
  else
#endif
    return ((double) clock())/CLOCKS_PER_SEC;
}
typedef complex<double> Complex;
inline double fabs(Complex x) { return fabs(x.real())+fabs(x.imag());}

//#include "sfem.hpp"
#define KN_CHECK
#include "RNM.hpp"
#include "GC.hpp"

// hack ---  F. Hecht -----
//  une fonction pour change un tableau de 
// complex en tableau de real 
//  Idee faire
// Une class qui tranforme une matrice complex en matric real
// et faire de transformateur de vecteur
KN_<double> C2R(KN_<complex<double> > & vc)
{ 
  assert(vc.step==1); 
  complex<double> * pc=vc; // pointeur du tableau
  double *pr = static_cast<double*>(static_cast<void*>(pc));
  return KN_<double>(pr,vc.N()*2);
}

KN_<complex<double> > R2C(KN_<double>  & vr)
{
  assert(vr.step==1 && vr.N() %2 ==0);
  double *pr =vr; // pointeur du tableau
   complex<double> * pc  = static_cast<complex<double>* >(static_cast<void*>(pr));
  return KN_<complex<double> >(pc,vr.N()/2);
}

const KN_<complex<double> > R2C(const KN_<double>  & vr)
{
  assert(vr.step==1 && vr.N() %2 ==0);
  double *pr =vr; // pointeur du tableau
   complex<double> * pc  = static_cast<complex<double>* >(static_cast<void*>(pr));
  return KN_<complex<double> >(pc,vr.N()/2);
}

//  une classe pour transforme une Matrice complex en Matrice  real 
// -----------------------------------------------------------------
template <class M>
class MatC2R : public VirtualMatrice<double> { public:
  typedef typename VirtualMatrice<double>::plusAx plusAx;
  //  typedef  VirtualMatrice<complex<double> > M;
  const M &m;
  MatC2R(const M &mm):m(mm) {}
  void addMatMul(const  KN_<double>  & x, KN_<double> & Ax) const {
    R2C(Ax) += m*R2C(x);
  }
  plusAx operator*(const KN<double> &  x) const {return plusAx(this,x);}
};

typedef double R; 

 
int main(int argc,char ** argv)
 {       
   typedef KN<Complex> Cn;
   typedef KN_<Complex> Cn_;
   typedef KNM<Complex> Cnm;
   typedef KNM<double> Rnm;
   typedef KN<double> Rn;
   typedef KN_<double> Rn_;
   typedef KNM_<Complex> Cnm_;
   typedef  MatC2R<Cnm> VMd;
   
   {
     
     int n=10;
     Cnm A(n,n),C(n,n),Id(n,n);
     Complex I(0,1);
     C=0;
     Id=0;
     Cn_ Aii(A,SubArray(n,0,n+1)); //  la diagonal de la matrice A sans copy	   
     Cn_ Cii(C,SubArray(n,0,n+1)); //  la diagonal de la matrice C sans copy	   
     Cn_ Idii(Id,SubArray(n,0,n+1)); //  la diagonal de la matrice Id sans copy	   
     for (int i=0;i<n;i++)
       for (int j=0;j<n;j++)
	 if(i==j)
	   Cii[i]= 1./(Aii[i]=2*n+1+i);
	 else
	   A(i,j) = i<j ? conj(Complex(1,1)) : Complex(1,1);
     Idii=1;
     cout << " A = \n" << A << endl;
     cout << " C = \n" << C << endl;
     Cn x(n),b(n),s(n),err(n);
     for (int i=0;i<n;i++) b[i]=Complex(i,(n-i));
     cout << "GradienConjugue preconditionne par la diagonale " <<  endl;
     x=Complex();
     GradienConjugue(A,C,b,x,n*2,1e-10);
     s = A*x ;
     cout << " solution : A*x= "  << s << endl; 
     err = s-b;
     cout << " err=" <<err.norm() << endl;
     assert(err.norm() < 1e-6); 
     cout << "GradienConjugue preconditionnee par la identity  " << endl;
     x=Complex();
     GradienConjugue(A,MatriceIdentite<Complex>(),b,x,n*3,1e-6);
     s = A*x ;
     cout << s << endl; 
     err = s-b;
     cout << " err 2= " <<err.norm()  << endl;
     assert(err.norm() < 1e-6); 
     // pour l'instant GMRES ne marche pas en complex 
	 //  mais avec un petit Hack qui transfome une 
	 // matrice complex en matrix real de taille double
	 // et de meme pour les vecteur tout marche bien
	 // CF. class MatC2R et class  C2R ---
	 //  remarque si la methede de converge pas 
	 //  pour resoudre $ A x = b  $ 
	 //   verifiez que $b \in Im A$, attention aux conditions aux limites
       R eps = 1e-6;
       int i50=50;  //  size of Krilov Space 
       Rnm H(i50+1,i50+1); // Warning gmres need a full matrix $(i50+1)^2$
       x=Complex();
       Rn_ rx=C2R(x);
       Rn_ rb=C2R(b);

	   int res=GMRES<VMd,Rn,VMd,Rnm,R,Rn_>(VMd(A),rx,rb,VMd(C),H,i50,n,eps);
       cout << " GMRES " << n << " eps = " << eps << " ok= 0==" << res << endl;
       s = A*x ;
       cout << " solution : A*x= "  << s << endl;
   
   }
   
   return 0;
 }

