#include <fstream>
#include <iostream>
#include <cassert>
#include <algorithm>
#include <map>
#include "gmres.hpp"
 using namespace std;
#include <time.h> 
inline double CPUtime(){
#ifdef SYSTIMES
  struct tms buf;
  if (times(&buf)!=-1)
    return ((double)buf.tms_utime+(double)buf.tms_stime)/(long) sysconf(_SC_CLK_TCK);
  else
#endif
    return ((double) clock())/CLOCKS_PER_SEC;
}

#define KN_CHECK
#include "RNM.hpp"
#include "GC.hpp"
typedef double R; 

class MatriceLaplacien1D: VirtualMatrice<R> { public:
 typedef  VirtualMatrice<R>::plusAx plusAx;
 MatriceLaplacien1D() {}; 
 void addMatMul(const KN_<R> & x, KN_<R> & Ax) const ;  
 plusAx operator*(const KN<R> &  x) const {return plusAx(this,x);} 
}; 
 
 void MatriceLaplacien1D::addMatMul(const KN_<R> & x, KN_<R> & Ax) const {  
  int n= x.N(),n_1=n-1;
  double h=1./(n_1),  d = 2/h, d1 = -1/h;
  Ax=0.;
  for (int i=1;i< n_1 ; i++)
    Ax[i] = (x[i-1] +x[i+1]) * d1 + x[i]*d  ; 
    
   //  CL 
   Ax[0]=x[0];
   Ax[n_1]=x[n_1];
  }

class MatriceMap: VirtualMatrice<R> { public:
 typedef  VirtualMatrice<R>::plusAx plusAx;
  typedef map<pair<int,int>,R> Map;
  
  const Map &m;
  MatriceMap(const Map & mm) :m(mm) {} 
 void addMatMul(const KN_<R> & x, KN_<R> & Ax) const ;  
 plusAx operator*(const KN<R> &  x) const {return plusAx(this,x);} 
}; 
 
 void MatriceMap::addMatMul(const KN_<R> & x, KN_<R> & Ax) const {  
   for ( Map::const_iterator k=m.begin(); k != m.end(); ++k)
     { 
       int i= k->first.first;
       int j= k->first.second;
       R  aij= k->second;
       Ax[i] += aij*x[j];
     }
  }

//  
void MatMul(const  map<pair<int,int>,R>  &A,const  map<pair<int,int>,R> &B,  map<pair<int,int>,R> &AB)
{
   // AB = A*B;
  typedef map<pair<int,int>,R> Map;
  for ( Map::const_iterator k=A.begin(); k != A.end(); ++k)
    {
      int i= k->first.first;
      int j= k->first.second;
       R  aij= k->second;
       //  pour toutes les termes   Bjk
       Map::const_iterator k1=B.upper_bound(make_pair(j+1,-1));// Finds the first element whose key greater than (j+1,0) 
       for (Map::const_iterator k0=B.upper_bound(make_pair(j,-1));k0 != k1;++k0)
	 {
	   int jj= k0->first.first;
	   int k= k0->first.second;
	   R  bjk= k0->second;
	   assert(jj == j);
	   AB[make_pair(i,k)] += aij*bjk;
	 }

    }
}
int main(int argc,char ** argv)
 {       
typedef KN<double> Rn;
typedef KN_<double> Rn_;
typedef KNM<double> Rnm;
typedef KNM_<double> Rnm_;
     
  { int n=10;
  Rnm A(n,n),C(n,n),Id(n,n);
  A=-1.;
  C=0.;
  Id=0.;
  Rn_ Aii(A,SubArray(n,0,n+1)); //  la diagonal de la matrice A sans copy	   
  Rn_ Cii(C,SubArray(n,0,n+1)); //  la diagonal de la matrice C sans copy	   
  Rn_ Idii(Id,SubArray(n,0,n+1)); //  la diagonal de la matrice Id sans copy	   
  for (int i=0;i<n;i++)
     Cii[i]= 1/(Aii[i]=n+i*i);
  Idii=1;
  cout << A;
  Rn x(n),b(n),s(n);
  for (int i=0;i<n;i++) b[i]=i;
  cout << "GradienConjugue preconditionne par la diagonale " <<  endl;
  x=0.;
  GradienConjugue(A,C,b,x,n,1e-10);
  s = A*x ;
  cout << " solution : A*x= "  << s << endl; 
  cout << "GradienConjugue preconditionnee par la identity  " << endl;
  x=0.;
  GradienConjugue(A,MatriceIdentite<R>(),b,x,n,1e-6);
  s = A*x ;
  cout << s << endl; 
  R eps = 1e-6;
  int i50=50;
  Rnm H(i50,i50);
  x=0.;
  //  int res=GMRES<Rnm,Rn,Rnm,Rnm,R,Rn_>(A,x,b,C,H,i50,n,eps);
  int res=GMRES<Rnm,Rn,Rnm,Rnm,R,Rn_>(A,x,b,C,H,i50,n,eps);
  cout << " GMRES " << n << " " << eps << " " << res << endl;
  s = A*x ;
  cout << " solution : A*x= "  << s << endl;
  }
  { 
  cout << "GradienConjugue laplacien 1D  par la identity  " << endl;
  int N=100;
  Rn b(N),x(N);
  R h= 1./(N-1);
  b= h;
  b[0]=0.;
  b[N-1]=0.;
  x=0.;
  R t0=CPUtime();
  GradienConjugue(MatriceLaplacien1D(),MatriceIdentite<R>() ,b,x,N,1e-5);
  cout << " Temps cpu = " << CPUtime() - t0<< "s" << endl;
  R err=0;
  for (int i=0;i<N;i++)
    {
    R xx=i*h;
    err= max(fabs(x[i]- (xx*(1-xx)/2)),err);
    }
  cout << "Fin err=" << err << endl;
  }

  {
    //  - Delta u = 1 
    int N = 100;
    int N1= N+1;
    double h = 1./N;
    double cc= -1;
    double aa= cc*-4/(h*h);
    double bb= cc/(h*h);
    map< pair<int,int> , double> A; 
    map< pair<int,int> , double> A2;
 
    for (int i0=1;i0<N;i0++)
      for (int i1=1;i1<N;i1++)
	{ 
	  int i = i0*N1+i1;
	  int jj[4] = { (i0-1)*N1+i1, (i0+1)*N1+i1, (i0)*N1+i1-1, (i0)*N1+i1+1 };
	  A[make_pair(i,i)] = aa;
	  for (int k=0;k<4;++k)
	    A[make_pair(i,jj[k])] = bb;
	}
    
    //    MatMul(A,A,A2);

    int n = N1*N1;
    Rn b(n),x(n);
    b =0.;
    for (int i0=1;i0<N;i0++)
      for (int i1=1;i1<N;i1++)
	{ 
	  int i = i0*N1+i1;
	  b[i]=1;
	}
    x=0.;
    R t0=CPUtime();
    GradienConjugue(MatriceMap(A),MatriceIdentite<R>() ,b,x,n,1e-5);
    cout << " Temps cpu = " << CPUtime() - t0<< "s" << endl;
    cout << " x.max = "<< x.max() << endl;
    ofstream ff("pp.gp");
    for (int i0=0;i0<=N;i0++)
      {
      for (int i1=0;i1<=N;i1++)
        {
          int i = i0*N1+i1;
          ff << i0 << " " << i1 << " " << x[i] << endl;
        }
      ff << endl ;
      }
    }
  return 0;
}
