
#include <cassert>
#include <cmath>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include "Mesh2d.hpp"
#include "RNM.hpp"
#include "GC.hpp"
#include "gmres.hpp"


int m=20, L=200;
R sigma1=0.2,sigma2=0.1, q=-0.25, r=0.02, K=100., dt=0.01, T=0.25;

R u0(const R2 & u){return Max(K-u.x-u.y,0.);} // right hand side
R g(const R2 & ){return 0;} // boundary condition

class MatLap2d: public VirtualMatrice<R> { public:
    //  Matrice $ \int \alpha u v + \beta \nabla u \nabla v $
    const  Mesh & Th;
  const R alpha,beta,gamma;
  typedef  VirtualMatrice<R>::plusAx plusAx;
  MatLap2d(const  Mesh & T,R a,R b, R c) : Th(T),alpha(a),beta(b), gamma(c) {};
  void addMatMul(const  KN_<R>  & x, KN_<R> & GDx) const {
      for (int k=0;k<Th.nt;k++)
	{
            const Triangle & K(Th[k]);
            int i0(Th(K[0])),i1(Th(K[1])),i2(Th(K[2])); // numero globaux des 3 sommets
            R2 H0(K.H(0)),H1(K.H(1)),H2(K.H(2));
            R x0=x[i0],x1=x[i1],x2=x[i2];
            R x012=x0+x1+x2; // la somme
            R2 gradx= H0*x0 + H1*x1 + H2*x2;

            R2 q0(K[0]), q1(K[1]), q3(K[2]), M( (q0+q1+q3)/3. ); //v[i0].x+v[i1].x+v[i2].x)/3

	    //la matrice Mij=\int_K w_i w_j
	    R m=K.area/12;
	    R Mx0 = (x012 + x0)*m;       //(2*x0 +   x1 + x2)*m;
	    R Mx1 = (x012 + x1)*m;       //(  x0 + 2*x1 + x2)*m;
	    R Mx2 = (x012 + x2)*m;       //(  x0 + x1 + 2*x2)*m;

	    //la matrice Bij=\int_K \mu * \nabla w_i * w_j
	    R2 mu((-r + (1./2)*q*sigma1*sigma2+pow(sigma1,2))*(M.x), (-r+ (1./2)*q*sigma1*sigma2+pow(sigma2,2))*(M.y));
            R bm=K.area/3;
	    R Bx0 =(mu,gradx)*bm;
	    R Bx1 =(mu,gradx)*bm;
	    R Bx2 =(mu,gradx)*bm;

	    //la matrice Aij=\int K \nabla w_i \nabla w_j + \int \nu w_i w_j  et \int \nu w_i w_j =Mij
	    //K matrice
	    R K00=(1./2)*pow(sigma1,2)*pow((M.x),2)*(K.area);
            R K01=(1./2)*q*sigma1*sigma2*(M.x)*(M.y)*(K.area);
	    R2 K0(K00,K01);
	    R K10=K01;
            R K11=(1./2)*pow(sigma2,2)*pow((M.y),2)*(K.area);
	    R2 K1(K10, K11);
	    R nu=r;
            //produit de matrice K par le vecteur \nabla w_i => (2x1)
            R2 KH0((K0,H0),(K1,H0));       //R2 KH0(K00*H0.x+K01*H0.y, K10*H0.x+K11*H0.y);
            R2 KH1((K0,H1),(K1,H1));       //R2 KH1(K00*H1.x+K01*H1.y, K10*H1.x+K11*H1.y);
            R2 KH2((K0,H2),(K1,H2));       //R2 KH2(K00*H2.x+K01*H2.y, K10*H2.x+K11*H2.y);

	    //produit scalaire
	    R Ax0=(gradx,KH0)+nu*Mx0;
	    R Ax1=(gradx,KH1)+nu*Mx1;
	    R Ax2=(gradx,KH2)+nu*Mx2;

            if ( K[0].lab!=2 &&  K[0].lab!=3 ) GDx[i0] +=alpha*Mx0+beta*Ax0+gamma*Bx0;
            if ( K[1].lab!=2 &&  K[1].lab!=3 ) GDx[i1] +=alpha*Mx1+beta*Ax1+gamma*Bx1;
            if ( K[2].lab!=2 &&  K[2].lab!=3 ) GDx[i2] +=alpha*Mx2+beta*Ax2+gamma*Bx2;
	} }
  plusAx operator*(const KN<R> &  x) const {return plusAx(this,x);}
};
//premier sembre (membre de Gauche)
//G[i0] += (Mx0+Ax0);
//G[i1] += (Mx1+Ax1);
//G[i0] += (Mx1+Ax1);

//second membre (membre de Droite)
//D[i0] += (Mx0-Bx0);
//D[i1] += (Mx1-Bx1);
//D[i2] += (Mx2-Bx2);

int main(int argc, char** argv )
{
  Mesh Th(argc > 1 ? argv[1] : "carre.msh");
  bool GC= !( argc > 2 );  // choix a la resolution du systeme lineaire
  cout << " GC = " << GC << endl;

  MatLap2d G(Th,1.,dt,0.); // Matrice de Masse  //delta t=0.001 (pas de temps =1)
  MatLap2d D(Th,1.,0.,-dt);  // Matrice du Laplacien
      
  MatriceIdentite<R> Id; // Matrice Identite
  KN<R> u0h(Th.nv);
  KN<R> b(Th.nv);  //  $ b[i] =  \int_\Omega  f_h w_i = M*fh  $
  KN<R> x(Th.nv);

  x=0.;  //  donne initial  avec les conditions aux limites.
  for (int k=0;k<Th.nt;k++)
    {
      Triangle & K(Th[k]);
      for (int i=0;i<3;i++)
	{
	  u0h[Th(k,i)] = u0(K[i]);
	  if(K[i].lab==2 || K[i].lab==3)  x[Th(K[i])] = g(K[i]);
	}
    }
  int res;
  for (int n=0; n*dt<=T+dt/100; n++)
    {  b = D*u0h;
  //  resolution du systeme lineaire
    if(GC)
    // attention      return 1 => convergence
    res= -1 + GradienConjugue(G,Id, b,x,Th.nv,1e-10);
  else
    {
      int m=20,max_iter=Th.nv;
      R tol=1e-6;
      KNM<R>  H(m,m);
      // attention      return 0  => convergence
      res = GMRES<VirtualMatrice<R>,KN<R>,VirtualMatrice<R>,KNM<R>,R,KN<R> >(G,x,b,Id,H,m,max_iter,tol);
      cout << " GMRES " << (res==0) << " tol =" << tol << " nb iter = " <<max_iter << " dim krilov : " << m << endl;
    }
    u0h=x;
    }

   if(res==0)
     {
       { // a file for gnuplot
	 ofstream plot("q4");
	 for(int it=0;it<Th.nt;it++)
	   plot << (R2) Th[it][0] << " " << x[Th(it,0)]<< endl
		<< (R2) Th[it][1] << " " << x[Th(it,1)] << endl
		<< (R2) Th[it][2] << " " << x[Th(it,2)] << endl
		<< (R2) Th[it][0] << " " << x[Th(it,0)] << endl << endl << endl;
       }
 	 for(int it=0;it<Th.nt;it++)
	   {if (Th[it][0].x==50 && Th[it][0].y==50)  cout<<"la valeur de u(50,50,T) est"<<"\t"<<x[Th(it,0)]<<endl; }
       {
	 ofstream f("x.sol");
	 f << x << endl;
       }
     }
   else cerr <<" Non convergence \n";
   return res;
}




