
#include <cassert>
#include <cmath>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include "Mesh2d.hpp"
#include "RNM.hpp"
#include "GC.hpp"
#include "gmres.hpp"

typedef double R;
typedef KNM<R> Rnm;
typedef KN<R> Rn;

R L = 200.;
R deltat=0.01;
R r,sigma1,sigma2,qee,KK,T;
char maille[20];


R f1(const R2 & F ){return Max(KK-F.x-F.y,0.);} // right hand side 
R g(const R2 & ){return 0;} // boundary condition
R u0(const R2 & ){return 0;} // initialization



class MatLap2d: public VirtualMatrice<R> { public:
    //  Matrice $ \int \alpha u v + \beta \nabla u \nabla v $
    const  Mesh & Th;
    const R alpha,beta,delta;
    typedef  VirtualMatrice<R>::plusAx plusAx;
    MatLap2d(const  Mesh & T,R a,R b,R c) : Th(T),alpha(a),beta(b), delta(c) {};

    void addMatMul(const  KN_<R>  & x, KN_<R> & Ax) const {
        for (int k=0;k<Th.nt;k++)
        {
            const Triangle & K=Th[k];
            int i0(Th(K[0])),i1(Th(K[1])),i2(Th(K[2])); // numero globaux des 3 sommets


            R2 H0(K.H(0)),H1(K.H(1)),H2(K.H(2));
            R x0=x[i0],x1=x[i1],x2=x[i2];
// cout <<"k="<<k<< ","<< "x0="<<x0<<" , "<< "x1="<<x1<<" , "<< "x2="<<x2<<" , "<<endl;
            R x012=x0+x1+x2; // la somme
            R2 gradx= H0*x0 + H1*x1 + H2*x2;
            
 	    R2 grav=(Th[k][0]+Th[k][1]+Th[k][2])/3.;
//	    définition de K (du probleme) et son calcul
  	    R2 K1,K2;
 	    (K1.x)=sigma1*sigma1*(grav.x)*(grav.x)/2.;
 	    (K1.y)=sigma1*sigma2*qee*(grav.x)*(grav.y)/2.;
 	    (K2.x)=(K1.y);
 	    (K2.y)=sigma2*sigma2*(grav.y)*(grav.y)/2.;
            R2 Kgrad0(K1.x*H0[0] + K2.x*H0[1],K1.y*H0[0] + K2.y*H0[1] );
            R2 Kgrad1(K1.x*H1[0] + K2.x*H1[1],K1.y*H1[0] + K2.y*H1[1] );
            R2 Kgrad2(K1.x*H2[0] + K2.x*H2[1],K1.y*H2[0] + K2.y*H2[1] );

//	    définition de mu et son calcul
	    R2 mu;
	    
	    mu.x=grav.x *(((-sigma1*sigma2*qee)/2.) - sigma1*sigma1 + r);
	    mu.y=grav.y *(((-sigma1*sigma2*qee)/2.) - sigma2*sigma2 + r);

 
	    //définition de r a été faite au début du programme
            R c1=(K.area*alpha/12.) , c2=K.area*beta , c3=delta*K.area/3.;
            if ((Th(i0).lab)!=3 && (Th(i0).lab)!=2) Ax[i0] +=  (x012+x0)*c1 +(gradx,Kgrad0)*c2 +(mu,gradx)*c3;
            if ((Th(i1).lab)!=3 && (Th(i1).lab)!=2) Ax[i1] +=  (x012+x1)*c1 +(gradx,Kgrad1)*c2 +(mu,gradx)*c3;
            if ((Th(i2).lab)!=3 && (Th(i2).lab)!=2) Ax[i2] +=  (x012+x2)*c1 +(gradx,Kgrad2)*c2 +(mu,gradx)*c3;
        }
    }
    plusAx operator*(const KN<R> &  x) const {return plusAx(this,x);}
};


    void CalculG(const Mesh & Th){ 
	ofstream ff("g.data");
	ff << Th.nv << "\t" << Th.nt << "\t" << Th.area << endl;
	for (int k=0;k<Th.nt;k++)
        {  R2 g=(Th[k][0]+Th[k][1]+Th[k][2])/3.;
	ff << g.x << "\t" << g.y << "\t" << k << endl;
	}}


//==============================================
//	Compte le nombre de double dans un fichier
//==============================================
int Nbl(const char *fichier)
{ifstream fic(fichier);
double c;
int compte=0;
while ( (fic >>c).good() ){			//lecture de chaque ligne
compte++;
}
fic.clear();
return compte;}
//==============================================




int main(int argc, char** argv )
{
ifstream ff("projet.txt");
ff>>sigma1;
ff>>sigma2;
ff>>qee;
ff>>r;
ff>>KK;
ff>>T;
ff>>maille;
ff.clear();

const char* maille1= maille;
Mesh Th(maille1);



//Calcul du centre de gravité de tous les triangles.
  CalculG(Th);
  int res;
  bool GC= !( argc > 2 );  // choix a la resolution du systeme lineaire 
  cout << " GC = " << GC << endl; 
//  MatLap2d M(Th,1.,0.,0.); // Matrice de Masse  meme matrice que dans le projet 2

  MatLap2d A(Th,1. + r*deltat,deltat,0.);  // Matrice du Laplacien
  MatLap2d B(Th,1.,0.,deltat);

  MatriceIdentite<R> Id; // Matrice Identite
  KN<R> fh(Th.nv);  //  $ fh[i] =  \Pi_h  f  $
  KN<R> b(Th.nv);  //  $ b[i] =  \int_\Omega  f_h w_i = M*fh  $
  
  KN<R> x(Th.nv); 
  
  x=0.;  //  donne initial  avec les conditions aux limites.
  for (int k=0;k<Th.nt;k++)
    {
      Triangle & K(Th[k]);
      for (int i=0;i<3;i++)
	{
	  fh[Th(k,i)] = f1(Th[k][i]);
	  int i0(Th(K[i]));
	  if( Th(i0).lab==2 || Th(i0).lab==3)  x[Th(K[i])] = g(Th[k][i]);
 	  else x[Th(K[i])] = f1(Th[k][i]);
	}
    }

// on commet une erreur plus grande si on passe par la convertion double => int
// int dtmax = int(T/deltat);
// for (int dt=0;dt<dtmax;dt++)
for (int n=0; n*deltat <= T+deltat/100; n++)
{

  b = B*fh;
  //  resolution du systeme lineaire 

  if(GC) 
    // attention      return 1 => convergence 
    res= -1 + GradienConjugue(A,Id, b,x,Th.nv,1e-10);       
  else
    {
      int m=20,max_iter=Th.nv;
      R tol=1e-6;
      KNM<R>  H(m,m);
      // attention      return 0  => convergence 
      res = GMRES<VirtualMatrice<R>,KN<R>,VirtualMatrice<R>,KNM<R>,R,KN<R> >(A,x,b,Id,H,m,max_iter,tol);
      cout << " GMRES " << (res==0) << " tol =" << tol << " nb iter = " << max_iter << " dim krilov : " << m << endl; 
    }
	fh=x;
}
   if(res==0)
     { 
       { // a file for gnuplot 
	 ofstream plot("plot");
	 for(int it=0;it<Th.nt;it++)
	   plot << (R2) Th[it][0] << " " << x[Th(it,0)]<< endl 
		<< (R2) Th[it][1] << " " << x[Th(it,1)] << endl 
		<< (R2) Th[it][2] << " " << x[Th(it,2)] << endl 
		<< (R2) Th[it][0] << " " << x[Th(it,0)] << endl << endl << endl; 
       }
	

	{
	 ofstream f2("uh.txt");
	 f2<< x << endl;
       }    
     }
   else cerr <<" Non convergence \n";



int choix;
retour:
cout <<endl<<endl<<endl<<endl<<endl<<endl<<endl<<endl<<endl<<endl<<endl<<endl<<endl<<endl<<endl<<endl<<endl<<endl;
cout <<"******************************************************"<<endl;
cout <<"**						    **"<<endl;
cout <<"**	  Interface pour effectuer les choix	    **"<<endl;
cout <<"**						    **"<<endl;
cout <<"**			Projet 2                    **"<<endl;
cout <<"**						    **"<<endl;
cout <<"**		Equation de Black&Scholes           **"<<endl;
cout <<"**			  en 2D                     **"<<endl;
cout <<"**						    **"<<endl;
cout <<"**			   par			    **"<<endl;
cout <<"**                                                  **"<<endl;
cout <<"**		      Dalissier Eric                **"<<endl;
cout <<"**						    **"<<endl;
cout <<"******************************************************"<<endl;
cout <<endl<<endl;
cout <<"  1: tracer la solution avec Gnuplot                  "<<endl;
cout <<"  2: tracer la solution Cpp affcher par FreeFem++     "<<endl;
cout <<"  3: tracer la solution BlackScholes2D avec FreeFem++ "<<endl;
cout <<"  4: tracer la solution BS4 avec FreeFem++            "<<endl;
cout <<"  5: Avoir la valeur en un point donné                "<<endl;
cout <<"  6: Derivee a l aide des differences finies par K    "<<endl;
cout <<"  7: Derivee a l aide des differences finies par q    "<<endl;
cout <<"  8: Derivee a l aide de DA			      "<<endl;
cout <<"  9: Voir la representation du domaine en OpenGL      "<<endl;
cout <<"  0: ne rien faire ( par defaut )                     "<<endl;
cout <<"Faite votre choix:";
cin>>choix;
if(choix==1) system("gnuplot trace.gnu");
if(choix==2) system("./FreeFem++ ruelnac.edp");
if(choix==3) system("./FreeFem++ BlackScholes2D.edp");
if(choix==4) system("./FreeFem++ BS4.edp");
if(choix==8) system("./DA/DA");
if(choix==9) system("./glplot carre.msh uh.txt 0.5");
if(choix==5) {
int vux, vuy;
cout <<"Donner le choix de x : ";
cin >> vux;
cout<<endl;
cout <<"Donner le choix de y : ";
cin >> vuy;
cout <<endl;
	int nbvect=0; 
	for ( int n=0;n<Th.nt ; n++){
	{
	if ( (Th[n][0].x==vux) && (Th[n][0].y==vuy) )
	{nbvect=n; }
	}}
 	cout << "u("<<vux<<","<<vuy<<","<<T<<")= "<< x[Th(nbvect,0)]<<endl<<endl;
}



if(choix==6) {
R K1,K2;
    cout << " Entrer le nouveau K (si possible different de K="<<KK<<") : ";
    cin >>K1;
    KN<R> x2(Th.nv);
    x2=x;
    K2=KK;
    KK=K1;
  MatLap2d A(Th,1. + r*deltat,deltat,0.);  // Matrice du Laplacien
  MatLap2d B(Th,1.,0.,deltat);
  MatriceIdentite<R> Id; // Matrice Identite
  KN<R> fh(Th.nv);  //  $ fh[i] =  \Pi_h  f  $
  KN<R> b(Th.nv);  //  $ b[i] =  \int_\Omega  f_h w_i = M*fh  $
  KN<R> x(Th.nv); 
  x=0.;  //  donne initial  avec les conditions aux limites.
  for (int k=0;k<Th.nt;k++)
    {
      Triangle & K(Th[k]);
      for (int i=0;i<3;i++)
	{
	  fh[Th(k,i)] = f1(Th[k][i]);
	  int i0(Th(K[i]));
	  if( Th(i0).lab==2 || Th(i0).lab==3)  x[Th(K[i])] = g(Th[k][i]);
 	  else x[Th(K[i])] = f1(Th[k][i]);
	}
    }
for (int n=0; n*deltat <= T+deltat/100; n++)
{
  b = B*fh;
  if(GC) 
    res= -1 + GradienConjugue(A,Id, b,x,Th.nv,1e-10);       
  else
    {
      int m=20,max_iter=Th.nv;
      R tol=1e-6;
      KNM<R>  H(m,m);
      // attention      return 0  => convergence 
      res = GMRES<VirtualMatrice<R>,KN<R>,VirtualMatrice<R>,KNM<R>,R,KN<R> >(A,x,b,Id,H,m,max_iter,tol);
    }
	fh=x;
}
if (K2>K1){     x2/= abs(K2-K1);
                x/=abs(K2-K1);
                x-= x2;
                x=-x;
}
else{     x2/= abs(K2-K1);
          x/=abs(K2-K1);
          x-= x2;}
    ofstream f4("dfk.txt");
    f4<< x << endl;
       { // a file for gnuplot 
	 ofstream plot("gnudfk");
	 for(int it=0;it<Th.nt;it++)
	   plot << (R2) Th[it][0] << " " << x[Th(it,0)]<< endl 
		<< (R2) Th[it][1] << " " << x[Th(it,1)] << endl 
		<< (R2) Th[it][2] << " " << x[Th(it,2)] << endl 
		<< (R2) Th[it][0] << " " << x[Th(it,0)] << endl << endl << endl; 
       }
system("gnuplot tracedfk.gnu");
}


if(choix==7) {
R q1,q2;
    cout << " Entrer le nouveau q (si possible different de q="<<qee<<") : ";
    cin >>q1;
    KN<R> x2(Th.nv);
    q2=qee;
    x2=x;
    qee=q1;
  MatLap2d A(Th,1. + r*deltat,deltat,0.);  // Matrice du Laplacien
  MatLap2d B(Th,1.,0.,deltat);
  MatriceIdentite<R> Id; // Matrice Identite
  KN<R> fh(Th.nv);  //  $ fh[i] =  \Pi_h  f  $
  KN<R> b(Th.nv);  //  $ b[i] =  \int_\Omega  f_h w_i = M*fh  $
  KN<R> x(Th.nv); 
  x=0.;  //  donne initial  avec les conditions aux limites.
  for (int k=0;k<Th.nt;k++)
    {
      Triangle & K(Th[k]);
      for (int i=0;i<3;i++)
	{
	  fh[Th(k,i)] = f1(Th[k][i]);
	  int i0(Th(K[i]));
	  if( Th(i0).lab==2 || Th(i0).lab==3)  x[Th(K[i])] = g(Th[k][i]);
 	  else x[Th(K[i])] = f1(Th[k][i]);
	}
    }
for (int n=0; n*deltat <= T+deltat/100; n++)
{
  b = B*fh;
  if(GC) 
    res= -1 + GradienConjugue(A,Id, b,x,Th.nv,1e-10);       
  else
    {
      int m=20,max_iter=Th.nv;
      R tol=1e-6;
      KNM<R>  H(m,m);
      // attention      return 0  => convergence 
      res = GMRES<VirtualMatrice<R>,KN<R>,VirtualMatrice<R>,KNM<R>,R,KN<R> >(A,x,b,Id,H,m,max_iter,tol);
    }
	fh=x;
}
if (q2>q1){     x2/= abs(q2-q1);
                x/=abs(q2-q1);
                x-= x2;
                x=-x;
}
else{     x2/= abs(q2-q1);
          x/=abs(q2-q1);
          x-= x2;}
    ofstream f5("dfq.txt");
    f5<< x << endl;

       { // a file for gnuplot 
	 ofstream plot("gnudfq");
	 for(int it=0;it<Th.nt;it++)
	   plot << (R2) Th[it][0] << " " << x[Th(it,0)]<< endl 
		<< (R2) Th[it][1] << " " << x[Th(it,1)] << endl 
		<< (R2) Th[it][2] << " " << x[Th(it,2)] << endl 
		<< (R2) Th[it][0] << " " << x[Th(it,0)] << endl << endl << endl; 
       }
system("gnuplot tracedfq.gnu");
}
if(choix==0){exit(0);}
goto retour;

   return res;
}


